//Project: Robot arm v0.1
//Homepage: www.HomoFaciens.de
//Author Norbert Heinz
//Version: 1.0
//Creation date: 24.02.2016
//This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//For a copy of the GNU General Public License see http://www.gnu.org/licenses/
//
//compile with gcc robot-arm.c -o robot-arm -lm
//For details see:
//http://www.HomoFaciens.de/technics-machines-robot-arm-v0-1_en_navion.htm

#include <stdio.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>



#define BAUDRATE B115200
//#define ARDUINOPORT "/dev/ttyUSB0"
#define ARDUINOPORT "/dev/ttyACM0"
#define FALSE 0
#define TRUE 1
#define PI 3.1415927


#define MOVE_PAUSE 20

#define SERVO_POS_01 1750
#define SERVO_POS_02 1700
#define SERVO_POS_03 1000
#define SERVO_POS_04 1800
#define SERVO_POS_05 1600

int fd = 0;
struct termios TermOpt;

int servoPos_01 = SERVO_POS_01;
int servoPos_02 = SERVO_POS_02;
int servoPos_03 = SERVO_POS_03;
int servoPos_04 = SERVO_POS_04;
int servoPos_05 = SERVO_POS_05;





//+++++++++++++++++++++++ Start readport +++++++++++++++++++++++++++++++
char  readport(void){
  int n;
  char buff;
  n = read(fd, &buff, 1);
  if(n > 0){
    return buff;
  }
  return 0;
}
//------------------------ End readport --------------------------------

//+++++++++++++++++++++++ Start sendport +++++++++++++++++++++++++++++++
void sendport(unsigned char ValueToSend){
  int n;

  n = write(fd, &ValueToSend, 1);

  if (n < 1){
    //while(readport() != 'r');
    //usleep(AdvanceRateArduino * 1.1);
    //n = write(fd, &ValueToSend, 1);
    //if (n < 1){
      printf("write() of value failed!\n");
    //}
  }
  else{
    while(readport() != 'r');
    while(readport() != 0);//flush read buffer
  }

}
//------------------------ End sendport --------------------------------

//+++++++++++++++++++++++ Start openport +++++++++++++++++++++++++++++++
void openport(void){

    fd = open(ARDUINOPORT, O_RDWR | O_NOCTTY | O_NDELAY);
    if (fd == -1)  {
      printf("init_serialport: Unable to open port\n");
    }

    if (tcgetattr(fd, &TermOpt) < 0) {
      printf("init_serialport: Couldn't get term attributes\n");
    }
    speed_t brate = BAUDRATE; // let you override switch below if needed

    cfsetispeed(&TermOpt, brate);
    cfsetospeed(&TermOpt, brate);

    // 8N1
    TermOpt.c_cflag &= ~PARENB;
    TermOpt.c_cflag &= ~CSTOPB;
    TermOpt.c_cflag &= ~CSIZE;
    TermOpt.c_cflag |= CS8;
    // no flow control
    TermOpt.c_cflag &= ~CRTSCTS;

    TermOpt.c_cflag |= CREAD | CLOCAL;  // turn on READ & ignore ctrl lines
    TermOpt.c_iflag &= ~(IXON | IXOFF | IXANY); // turn off s/w flow ctrl

    TermOpt.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // make raw
    TermOpt.c_oflag &= ~OPOST; // make raw

    // see: http://unixwiz.net/techtips/termios-vmin-vtime.html
    TermOpt.c_cc[VMIN]  = 0;
    TermOpt.c_cc[VTIME] = 20;

    if( tcsetattr(fd, TCSANOW, &TermOpt) < 0) {
      printf("init_serialport: Couldn't set term attributes\n");
    }

}
//------------------------ End openport --------------------------------

//+++++++++++++++++++++++ Start storeNewCoordinates ++++++++++++++++++++
void storeNewCoordinates(){
  FILE *storeFile;
  char TextLine[30000];

  if((storeFile=fopen("current-coordinates.tmp", "w"))==NULL){
    printf("Can't open file 'stored-coordinates.txt'!\n");
  }
  else{
    sprintf(TextLine, "%05d %05d %05d %05d %05d\n", servoPos_01, servoPos_02, servoPos_03, servoPos_04, servoPos_05);
    fwrite(TextLine, 1, strlen(TextLine), storeFile);
    fclose(storeFile);
    system("mv current-coordinates.tmp current-coordinates.txt");
  }
}

//------------------------ End storeNewCoordinates ---------------------


void gotoCoordinates(int servoPos_01_new, int servoPos_02_new, int servoPos_03_new, int servoPos_04_new, int servoPos_05_new){


  while(servoPos_01_new != servoPos_01 || servoPos_02_new != servoPos_02 || servoPos_03_new != servoPos_03 || servoPos_04_new != servoPos_04 || servoPos_05_new != servoPos_05){
    if(servoPos_01_new > servoPos_01){
      sendport('a');
      servoPos_01++;
      usleep(MOVE_PAUSE);
    }
    if(servoPos_01_new < servoPos_01){
      sendport('b');
      servoPos_01--;
      usleep(MOVE_PAUSE);
    }

    if(servoPos_02_new > servoPos_02){
      sendport('c');
      servoPos_02++;
      usleep(MOVE_PAUSE);
    }
    if(servoPos_02_new < servoPos_02){
      sendport('d');
      servoPos_02--;
      usleep(MOVE_PAUSE);
    }

    if(servoPos_03_new > servoPos_03){
      sendport('e');
      servoPos_03++;
      usleep(MOVE_PAUSE);
    }
    if(servoPos_03_new < servoPos_03){
      sendport('f');
      servoPos_03--;
      usleep(MOVE_PAUSE);
    }

    if(servoPos_04_new > servoPos_04){
      sendport('g');
      servoPos_04++;
      usleep(MOVE_PAUSE);
    }
    if(servoPos_04_new < servoPos_04){
      sendport('h');
      servoPos_04--;
      usleep(MOVE_PAUSE);
    }

    if(servoPos_05_new > servoPos_05){
      sendport('i');
      servoPos_05++;
      usleep(MOVE_PAUSE);
    }
    if(servoPos_05_new < servoPos_05){
      sendport('j');
      servoPos_05--;
      usleep(MOVE_PAUSE);
    }
  }

}

//######################################################################
//################## Main ##############################################
//######################################################################

int main(int argc, char **argv){

  int i;
  FILE *storeFile;
  FILE *commandFile;
  char TextLine[30000];
  char commandReceived, commandReceived2;
  long MoveLength = 1;


  openport();

  printf("\n\nWaiting for 'X' from Arduino (Arduino pluged in?)...\n");

  //Wait for 'X' from Arduino
  while(readport() != 'X');

  printf("...connected to Arduino.\n");


  storeNewCoordinates();

  while(1){
    commandReceived = 0;

    if((commandFile=fopen("nextCommand.txt", "rb"))!=NULL){
      fread(&commandReceived, 1, 1, commandFile);
      fread(&commandReceived2, 1, 1, commandFile);
      fclose(commandFile);
      system("rm nextCommand.txt");
      MoveLength = 1;
      if(commandReceived2 == 'a'){
        MoveLength = 1;
      }
      if(commandReceived2 == 'b'){
        MoveLength = 10;
      }
      if(commandReceived2 == 'c'){
        MoveLength = 100;
      }
      printf("%c received, MoveLength = %ld\n", commandReceived, MoveLength);
    }


    if(commandReceived == 's'){
      if((storeFile=fopen("stored-coordinates.txt", "a"))==NULL){
        printf("Can't open file 'stored-coordinates.txt'!\n");
      }
      else{
        sprintf(TextLine, "%05d %05d %05d %05d %05d\n", servoPos_01, servoPos_02, servoPos_03, servoPos_04, servoPos_05);
        fwrite(TextLine, 1, strlen(TextLine), storeFile);
        printf(TextLine, "%05d %05d %05d %05d %05d stored.\n", servoPos_01, servoPos_02, servoPos_03, servoPos_04, servoPos_05);
        fclose(storeFile);
      }
    }


    if(commandReceived == 'D'){//Draw a square
      double L1 = 130.0;
      double L2 = 240.0;
      double L3, x, y, alpha, beta, delta, diffPos_01, diffPos_02;

      //Initial position of robot arm
      gotoCoordinates(SERVO_POS_01, SERVO_POS_02 - 1000, SERVO_POS_03, SERVO_POS_04, servoPos_05);
      sleep(1);
      gotoCoordinates(SERVO_POS_01, SERVO_POS_02 - 1000, SERVO_POS_03, SERVO_POS_04 + 700, servoPos_05);
      sleep(1);
      y = 0.0;
      x = 0.0;

      for(i = 0; i < 200.0; i++){
        if(i < 50){
          x+=1.0;
        }
        if(i > 49 && i < 100){
          y+=1.0;
        }
        if(i > 99 && i < 150){
          x-=1.0;
        }
        if(i > 149 && i < 200){
          y-=1.0;
        }
        L3 = sqrt(pow(L2-y, 2.0) + pow(L1 + x, 2.0));
        alpha = acos((pow(L1, 2.0) + pow(L3, 2.0) - pow(L2, 2.0))/(2 * L1 * L3));
        delta = asin((L1 + x)/L3);
        alpha = alpha - (PI/2.0 - delta);
        beta = acos((pow(L1, 2.0) + pow(L2, 2.0) - pow(L3, 2.0))/(2 * L1 * L2));
        diffPos_01 = 1000.0 * alpha / (PI/2.0);
        diffPos_02 = 1000.0 * beta / (PI/2.0) - 1000.0;
        gotoCoordinates(SERVO_POS_01 + (int)diffPos_01, SERVO_POS_02 - 1000 + (int)diffPos_02, SERVO_POS_03, SERVO_POS_04 + 700, servoPos_05);
        usleep(10000);
        printf("% 3d.)SERVO_POS_01 + diffPos_01 = %d, SERVO_POS_02 - 1000 + diffPos_02 = %d\n", i, SERVO_POS_01 + (int)diffPos_01, SERVO_POS_02 - 1000 + (int)diffPos_02);
      }
      sleep(1);
      gotoCoordinates(SERVO_POS_01 + (int)diffPos_01, SERVO_POS_02 - 1000 + (int)diffPos_02, SERVO_POS_03, SERVO_POS_04, servoPos_05);
      sleep(1);
      gotoCoordinates(SERVO_POS_01, SERVO_POS_02, SERVO_POS_03, SERVO_POS_04, servoPos_05);
      sleep(20);
    }


    if(commandReceived == 'r'){
      if((storeFile=fopen("stored-coordinates.txt", "r"))==NULL){
        printf("Can't open file 'stored-coordinates.txt'!\n");
      }
      else{
        int servoPos_01_new, servoPos_02_new, servoPos_03_new, servoPos_04_new, servoPos_05_new;
        int stopRun = 0;
        while(!feof(storeFile)){
          TextLine[0]='\0';
          fread(TextLine, 1, 6, storeFile);
          TextLine[6]='\0';
          servoPos_01_new = atoi(TextLine);
          if(servoPos_01_new < 500 || servoPos_01_new > 2500){
            stopRun = 1;
          }

          TextLine[0]='\0';
          fread(TextLine, 1, 6, storeFile);
          TextLine[6]='\0';
          servoPos_02_new = atoi(TextLine);
          if(servoPos_02_new < 500 || servoPos_02_new > 2500){
            stopRun = 1;
          }

          TextLine[0]='\0';
          fread(TextLine, 1, 6, storeFile);
          TextLine[6]='\0';
          servoPos_03_new = atoi(TextLine);
          if(servoPos_03_new < 500 || servoPos_03_new > 2500){
            stopRun = 1;
          }

          TextLine[0]='\0';
          fread(TextLine, 1, 6, storeFile);
          TextLine[6]='\0';
          servoPos_04_new = atoi(TextLine);
          if(servoPos_04_new < 500 || servoPos_04_new > 2500){
            stopRun = 1;
          }

          TextLine[0]='\0';
          fread(TextLine, 1, 6, storeFile);
          TextLine[6]='\0';
          servoPos_05_new = atoi(TextLine);
          if(servoPos_05_new < 500 || servoPos_05_new > 2500){
            stopRun = 1;
          }

          if(stopRun == 0){
            printf("Going to %05d %05d %05d %05d %05d\n", servoPos_01_new, servoPos_02_new, servoPos_03_new, servoPos_04_new, servoPos_05_new);
          }

          if(stopRun == 0){
            gotoCoordinates(servoPos_01_new, servoPos_02_new, servoPos_03_new, servoPos_04_new, servoPos_05_new);
            sleep(1);
          }
        }//while(!feof(storeFile)){
      }
    }//if(commandReceived == 'r'){

    if(commandReceived == 'a'){
      for(i=0; i<MoveLength; i++){
        sendport('a');
        usleep(MOVE_PAUSE);
      }
      servoPos_01 += MoveLength;
      if(servoPos_01 > 2500){
        servoPos_01 = 2500;
      }
    }
    if(commandReceived == 'b'){
      for(i=0; i<MoveLength; i++){
        sendport('b');
        usleep(MOVE_PAUSE);
      }
      servoPos_01 -= MoveLength;
      if(servoPos_01 < 500){
        servoPos_01 = 500;
      }
    }

    if(commandReceived == 'c'){
      for(i=0; i<MoveLength; i++){
        sendport('c');
        usleep(MOVE_PAUSE);
      }
      servoPos_02 += MoveLength;
      if(servoPos_02 > 2500){
        servoPos_02 = 2500;
      }
    }
    if(commandReceived == 'd'){
      for(i=0; i<MoveLength; i++){
        sendport('d');
        usleep(MOVE_PAUSE);
      }
      servoPos_02 -= MoveLength;
      if(servoPos_02 < 500){
        servoPos_02 = 500;
      }
    }

    if(commandReceived == 'e'){
      for(i=0; i<MoveLength; i++){
        sendport('e');
        usleep(MOVE_PAUSE);
      }
      servoPos_03 += MoveLength;
      if(servoPos_03 > 2500){
        servoPos_03 = 2500;
      }
    }
    if(commandReceived == 'f'){
      for(i=0; i<MoveLength; i++){
        sendport('f');
        usleep(MOVE_PAUSE);
      }
      servoPos_03 -= MoveLength;
      if(servoPos_03 < 500){
        servoPos_03 = 500;
      }
    }

    if(commandReceived == 'g'){
      for(i=0; i<MoveLength; i++){
        sendport('g');
        usleep(MOVE_PAUSE);
      }
      servoPos_04 += MoveLength;
      if(servoPos_04 > 2500){
        servoPos_04 = 2500;
      }
    }
    if(commandReceived == 'h'){
      for(i=0; i<MoveLength; i++){
        sendport('h');
        usleep(MOVE_PAUSE);
      }
      servoPos_04 -= MoveLength;
      if(servoPos_04 < 500){
        servoPos_04 = 500;
      }
    }

    if(commandReceived == 'i'){
      for(i=0; i<MoveLength; i++){
        sendport('i');
        usleep(MOVE_PAUSE);
      }
      servoPos_05 += MoveLength;
      if(servoPos_05 > 2500){
        servoPos_05 = 2500;
      }
    }
    if(commandReceived == 'j'){
      for(i=0; i<MoveLength; i++){
        sendport('j');
        usleep(MOVE_PAUSE);
      }
      servoPos_05 -= MoveLength;
      if(servoPos_05 < 500){
        servoPos_05 = 500;
      }
    }


    if(commandReceived == 27){
      exit(0);
    }

    if(commandReceived != 0){
      storeNewCoordinates();
    }
  }

  return 0;
}

